/*----------------------------------------------------------------------------*

	clock.c - CMOS Clock Manipulation
		Reworked for 2.1.x(2.2.x) by Alex Buell, 
			alex.buell@tahallah.demon.co.uk

	September 1998
	
 ----------------------------------------------------------------------------- 

 	clock [-u] -r	read cmos clock
	clock [-u] -w	write cmos clock from system time
	clock [-u] -s	set system time from cmos clock
	clock [-u] -a	set system time from cmos clock, adjust the time to
			correct for systematic error, and put it back to the cmos.
 			-u indicates cmos clock is kept in universal time
	clock -v	prints the version.

 ----------------------------------------------------------------------------- 

	HISTORY

	1.0	- by Charles Hedrick, hedrick@cs.rutgers.edu, April 1992
	1.1	- Modifed by Rob Hooft, hooft@chem.ruu.nl, November 1992
	1.2	- Patched by Rob Hooft with input from Harald Koening.
	1.3	- Modified by Alan Modra, alan@spri.levels.unisa.edu.au
	1.4	- Modifed for >2000 by Alan Modra, February 1995
	1.4b	- Sparc mods by Tom Dyas, tdyas@eden.rutgers.edu, August 1996
	1.5	- Mods for 2.1.x(2.2.x) by Alex Buell, September 1998
	1.5.1	- Mods to bring the 2.0.x code back in, by Alex Buell, October 1998

-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>

#include <linux/version.h>
#define KERNEL_VER(major, minor, sublevel) (((major) << 16) + ((minor) << 8) + (sublevel))

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,0)
#include <sys/ioctl.h>
#include <linux/mc146818rtc.h>
#else
#include <asm/rtc.h>
#endif 

/* Read the README file for further details */

/* Here the information for time adjustments is kept. */
#define ADJPATH "/etc/adjtime"

/* used for debugging the code. */
/* #define KEEP_OFF */

/* Globals */
int readit = 0;
int adjustit = 0;
int writeit = 0;
int setit = 0;
int universal = 0;
int debug = 0;

int rtc_fd;

volatile void usage () 
{
	fprintf (stderr,"clock [-u] -r|w|s|a|v\n"
			"  r: read and print CMOS clock\n"
			"  w: write CMOS clock from system time\n"
			"  s: set system time from CMOS clock\n"
			"  a: get system time and adjust CMOS clock\n"
			"  u: CMOS clock is in universal time\n"
			"  v: print version (" VERSION ") and exit\n");
	exit (1);
}


void rtc_init ()
{
	rtc_fd = open ("/dev/rtc", 2);
	if (rtc_fd < 0)
	{
		perror ("Unable to open /dev/rtc read/write ");
		exit (1);
	}
}

int main (int argc, char **argv, char **envp)
{
	struct tm tm;
	time_t systime;
	time_t last_time;
	char arg;
	double factor;
	double not_adjusted;
	int adjustment = 0;
	unsigned char save_control, save_freq_select;
	struct rtc_time rtc_tm;

	while ((arg = getopt (argc, argv, "rwsuaDv")) != -1)
	{
		switch (arg)
		{
			case 'r':
				readit = 1;
				break;
			case 'w':
				writeit = 1;
				break;
			case 's':
				setit = 1;
				break;
			case 'u':
				universal = 1;
				break;
			case 'a':
				adjustit = 1;
				break;
			case 'D':
				debug = 1;
				break;
			case 'v':
				fprintf( stderr, "clock " VERSION "\n" );
				exit(0);
			default:
				usage ();
		}
	}

	if (readit + writeit + setit + adjustit > 1)
		usage ();			/* only allow one of these */

	if (!(readit | writeit | setit | adjustit))	/* default to read */
		readit = 1;

	rtc_init ();

	if (adjustit)
	{				/* Read adjustment parameters first */
		FILE *adj;

		if ((adj = fopen (ADJPATH, "r")) == NULL)
		{
			perror (ADJPATH);
			exit (2);
		}

		if (fscanf (adj, "%lf %d %lf", &factor, &last_time, &not_adjusted) < 0)
		{
			perror (ADJPATH);
			exit (2);
		}
		fclose (adj);

		if (debug) 
			printf ("Last adjustment done at %d seconds after 1/1/1970\n", last_time);
	}

	if (readit || setit || adjustit)
	{
		int i;

/* The purpose of the "do" loop is called "low-risk programming" */
/* In theory it should never run more than once */

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,0)
		if (ioctl(rtc_fd, RTC_RD_TIME, &rtc_tm) < 0)
		{
			perror("ioctl(RTC_RD_TIME)");
			exit (2);
		}
      
		tm.tm_sec = rtc_tm.tm_sec;
		tm.tm_min = rtc_tm.tm_min;
		tm.tm_hour = rtc_tm.tm_hour;
		tm.tm_wday = rtc_tm.tm_wday;
		tm.tm_mday = rtc_tm.tm_mday;
		tm.tm_mon = rtc_tm.tm_mon;
		tm.tm_year = rtc_tm.tm_year;
		tm.tm_isdst = -1;		/* don't know whether it's daylight */
#else
		if (ioctl(rtc_fd, RTCGET, &rtc_tm) < 0)
		{
			perror("ioctl(RTCGET)");
			exit (2);
		}

		tm.tm_sec = rtc_tm.sec;
		tm.tm_min = rtc_tm.min;
		tm.tm_hour = rtc_tm.hour;
		tm.tm_wday = rtc_tm.dow - 1;
		tm.tm_mday = rtc_tm.dom;
		tm.tm_mon = rtc_tm.month - 1;
		tm.tm_year = rtc_tm.year - 1900;
		tm.tm_isdst = -1; /* as above */
#endif
	}

	if (readit || setit || adjustit)
	{

/*
 * mktime() assumes we're giving it local time.  If the CMOS clock
 * is in GMT, we have to set up TZ so mktime knows it.  tzset() gets
 * called implicitly by the time code, but only the first time.  When
 * changing the environment variable, better call tzset() explicitly.
 */

		if (universal)
		{
			char *zone;
			zone = (char *)getenv("TZ");	/* save original time zone */
			(void)putenv("TZ=");
			tzset ();
			systime = mktime (&tm);

			/* now put back the original zone */
			if(zone)
			{
				char *zonebuf;
				zonebuf = malloc(strlen(zone) + 4);
				strcpy(zonebuf, "TZ=");
				strcpy(zonebuf+3, zone);
				putenv(zonebuf);
				free(zonebuf);
			}
			else
			{
	      			putenv("TZ"); /* no timezone? clear it. */
	 		}
 
			tzset ();
		}
		else
		{
			systime = mktime(&tm);
		}

		if (debug) 
			printf ("Number of seconds since 1/1/1970 is %d\n", systime);
	}

	if (readit)
		printf ("%s", ctime (&systime ));

	if (setit || adjustit)
	{
		struct timeval tv;
		struct timezone tz;

	/* program is designed to run setuid, be secure! */

		if (getuid() != 0)
		{			
			fprintf (stderr, "Sorry, must be root to set or adjust time\n");
			exit (2);
		}

		if (adjustit)
		{			/* the actual adjustment */
			double exact_adjustment;

			exact_adjustment = ((double) (systime - last_time)) * factor / (24 * 60 * 60) + not_adjusted;

			if(exact_adjustment > 0)
				adjustment = (int)(exact_adjustment + 0.5);
			else
				adjustment = (int)(exact_adjustment - 0.5);

			not_adjusted = exact_adjustment - (double)adjustment;
			systime += adjustment;

			if (debug) 
			{
				printf("Time since last adjustment is %d seconds\n", (int) (systime - last_time));
				printf("Adjusting time by %d seconds\n", adjustment);
				printf("Remaining adjustment is %.3f seconds\n", not_adjusted);
			}
		}

#ifndef KEEP_OFF
		tv.tv_sec = systime;
		tv.tv_usec = 0;
		tz.tz_minuteswest = timezone / 60;
		tz.tz_dsttime = daylight;

		if (settimeofday (&tv, &tz) != 0)
       		{
			fprintf (stderr, "Unable to set time -- probably you are not root\n");
			exit (1);
		}
      
		if (debug) 
		{
			printf("Called settimeofday:\n");
			printf("\ttv.tv_sec = %ld, tv.tv_usec = %ld\n",
			tv.tv_sec, tv.tv_usec );
			printf("\ttz.tz_minuteswest = %d, tz.tz_dsttime = %d\n", tz.tz_minuteswest, tz.tz_dsttime );
		}
#endif
	}
  
	if(writeit || (adjustit && adjustment != 0))
	{
		struct tm *tmp;
		systime = time (NULL);

		if (universal)
			tmp = gmtime(&systime);
		else
			tmp = localtime(&systime);

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,0)
		rtc_tm.tm_sec = tmp->tm_sec;
		rtc_tm.tm_min = tmp->tm_min;
		rtc_tm.tm_hour = tmp->tm_hour;
		rtc_tm.tm_wday = tmp->tm_wday;
		rtc_tm.tm_mday = tmp->tm_mday;
		rtc_tm.tm_mon = tmp->tm_mon;
		rtc_tm.tm_year = tmp->tm_year;

		if(ioctl(rtc_fd, RTC_SET_TIME, &rtc_tm) < 0)
		{
			perror("ioctl(RTC_SET_TIME)");
			exit (2);
		}
#else
		rtc_tm.sec = tmp->tm_sec;
		rtc_tm.min = tmp->tm_min;
		rtc_tm.hour = tmp->tm_hour;
		rtc_tm.dow = tmp->tm_wday + 1;
		rtc_tm.dom = tmp->tm_mday;
		rtc_tm.month = tmp->tm_mon + 1;
		rtc_tm.year = tmp->tm_year + 1900;

		if (ioctl(rtc_fd, RTCSET, &rtc_tm) < 0)
		{
			perror("ioctl(RTCSET)");
			exit(2);
		}
#endif /* Linux 2.0.x */	

		if (debug) 
			printf ("Set to : %02d:%02d:%02d\n", tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
	}
	else
		if (debug) 
			printf ("CMOS clock unchanged.\n");

	/* Save data for next 'adjustit' call */
	if (adjustit)
	{
		FILE *adj;
		if((adj = fopen (ADJPATH, "w")) == NULL)
		{
			perror(ADJPATH);
			exit (2);
		}
		
		fprintf(adj, "%f %d %f\n", factor, systime, not_adjusted);
		fclose (adj);
	}

	exit (0);
}
